#include <ros/ros.h>
#include <iostream>
#include <math.h>

#include <std_msgs/Float32.h>
#include <std_msgs/Float64.h>
#include <std_msgs/Bool.h>
#include <geometry_msgs/Twist.h>

using namespace std;



//Intensities of the 4 motors. To be accessed at other places of the program
float int_fl;
float int_fr;
float int_bl;
float int_br;

float bat_lev;

float ctl_FL;
float ctl_FR;
float ctl_BL;
float ctl_BR;

float speedX;       //linear speed
float speedZ;       //angular speed

int ind = 0;
int indStDev = 0;
bool ArrayCompleted = false;    //this variable will tell when the array is full and computations can start
bool stdDevArrCompleted = false;

//powerVariance variables
const int buffSize = 20;            //this variable determines the length of the buffer, how many powers will be saved for computing the variance
float varArray[buffSize] = {};      //array of variance
const int buffStdDev = 20;          //size of the buffer of standard deviations
float stDevArray[buffStdDev] = {};  //creation of the array for the stdDev

bool cleanFirstArray = true;        //this will empty the first array of intensities since it is noise due to the stabilization of the PID signal


/* These callbacks will be used in order to extract the data sent
 * by the topics on ROS
 */
void I_flCallback(const std_msgs::Float64& msg){
    int_fl = msg.data;
}

void I_frCallback(const std_msgs::Float64& msg){
    int_fr = msg.data;
}

void I_blCallback(const std_msgs::Float64& msg){
    int_bl = msg.data;
}

void I_brCallback(const std_msgs::Float64& msg){
    int_br = msg.data;
}

void batCallback(const std_msgs::Float64& msg){
    bat_lev = msg.data;
}

void ctl_flCallback(const std_msgs::Float32::ConstPtr& msg){
    ctl_FL = msg->data;
}

void ctl_frCallback(const std_msgs::Float32::ConstPtr& msg){
    ctl_FR = msg->data;
}

void ctl_blCallback(const std_msgs::Float32::ConstPtr& msg){
    ctl_BL = msg->data;
}

void ctl_brCallback(const std_msgs::Float32::ConstPtr& msg){
    ctl_BR = msg->data;
}

void vel_Callback(const geometry_msgs::Twist::ConstPtr& msg){
    speedX = msg->linear.x;
    speedZ = msg->angular.z;
}

float* computePowers(){
//computing the powers for the 4 different wheels

    float* p = new float[4];

    p[0] = float((int_fl/1000.0)*bat_lev*ctl_FL);   //FL
    p[1] = float((int_fr/1000.0)*bat_lev*ctl_FR);   //FR
    p[2] = float((int_bl/1000.0)*bat_lev*ctl_BL);   //BL
    p[3] = float((int_br/1000.0)*bat_lev*ctl_BR);   //BR

    return p;
}



bool powerVariance(float* Pw){
/*
 * This function will provide information about the variance of the powers of the different wheels.
 * The mean of the power of the 4 wheels will be computed and stored in an array of X positions.
 * This wants to represent the variance of the power during time.
 *
 * If the Rate is set to 10 and there are 20 positions on the array it represents 2 seconds of the
 * reality.
 *
 * If a big variance is observed for a certain terrain it will be necessary to do a test to be sure
 * that the robot has changed or not the terrain.
 *
 * This variation will be determined by the standard deviation (given by the squared root of the variance)
 * If the stdDev is changing and not constant during a certain period of time it means that probably
 * the terrain has changed.
 *
 * */

    float Pw4w = 0;                     //Power of the four wheels
    float meanPw;
    float variance = 0;
    float stdDev = 0;
    bool testLaunch = false;            //this variable says if it is needed to check the terrain
    float meanStdDev = 0;
    int outliers = 0;                   //number of stdDev that will be outside the range. This will mark when the terrain has changed
    float thrMax = 1.3;                 //threshold for the stdDev. Right now it is using a 30%. If a value is outside this thr it will be considered an outlier
    float thrMin = 0.7;
    int maxOutliers = 7;                //set the maximum outliers allowed

    Pw4w = ( Pw[0] + Pw[1] + Pw[2] + Pw[3] ) / 4;   //power of the 4 wheels


    if (speedX == 0){
        /* If the linear speed is equal to 0 it means that the test has to begin
         * again. In order to do that it is necessary to reset the arrays to 0 and
         * to set again the indices to 0. It is also necessary to set the flags of
         * the arrays to 'false' so the program knows that it has to complete again
         * them before doing new computations.
         */

        for(int jj=0; jj<buffSize; jj++){
            varArray[jj] = 0;
        }
        for(int hh=0; hh<buffStdDev; hh++){
            stDevArray[hh] = 0;
        }
        ind = 0;
        indStDev = 0;
        ArrayCompleted = false;
        stdDevArrCompleted = false;
        cleanFirstArray = true;
    }

    varArray[ind] = Pw4w;       //allocating the mean of the 4 powers in the array
    ind++;


    if (ArrayCompleted && speedZ==0 && speedX>0){       //limit the program to work only when the robot goes forward

        for (int i=0; i<buffSize; i++){     //compute the mean power
            meanPw = meanPw + varArray[i];
        }
        meanPw = meanPw / buffSize;

        for (int j=0; j<buffSize; j++){     //compute the variance
            variance = variance + pow((varArray[j] - meanPw), 2.0);
        }
        variance = variance / buffSize;   //averaging the variance

        stdDev = sqrt(variance);            //compute the standard deviation

        stDevArray[indStDev] = stdDev;      //allocating the standard deviations in the array
        indStDev++;

        if (stdDevArrCompleted){

            for (int k=0; k<buffStdDev; k++){               //compute the mean of the standard deviation
                meanStdDev = meanStdDev + stDevArray[k];
            }
            meanStdDev = meanStdDev / buffStdDev;

            for (int l=0; l<buffStdDev; l++){               //checking the outliers from the mean
                if (stDevArray[l]>meanStdDev*thrMax || stDevArray[l]<meanStdDev*thrMin){
                    outliers++;
                }
                //cout << stDevArray[l] << " ";
            }
            //cout << endl;
            cout << "Outliers: " << outliers << "  " << "Current standard deviation: " << stdDev << " FL: " << Pw[0] << " FR: " << Pw[1] << " BL: " << Pw[2] << " BR: " << Pw[3] << endl;

            if (outliers > maxOutliers){
                testLaunch = true;
                cout << "Change of terrain detected!!!" << "  Outliers: " << outliers << endl;
            }
        }


    }


    //if the index is greater than 19 it means that we are outside of the array
    //so it must be set to 0 again
    if (ind == buffSize){
        ind = 0;
        ArrayCompleted = true;
    }

    if (indStDev == buffStdDev){
        indStDev = 0;
        stdDevArrCompleted = true;

        if(cleanFirstArray==true){              //first set of stdDev suffer from stabilization of PID signal so they must be "destroied"

            for(int jjj=0; jjj<buffSize; jjj++){
                varArray[jjj] = 0;
            }
            for(int hhh=0; hhh<buffStdDev; hhh++){
                stDevArray[hhh] = 0;
            }
            ind = 0;
            indStDev = 0;

            ArrayCompleted = false;
            stdDevArrCompleted = false;
            cleanFirstArray = false;
            //cout << "Reset the parameters" << endl;
        }
    }

return testLaunch;

}


int main(int argc, char** argv){

    ros::init(argc, argv, "check_terrain");
    ros::NodeHandle m;


    /* Declaration of all the subscribers that will be needed during the
     * all the computations of this node*/

    ros::Subscriber I_fl = m.subscribe("Current_Front_Left", 100, I_flCallback);
    ros::Subscriber I_fr = m.subscribe("Current_Front_Right", 100, I_frCallback);
    ros::Subscriber I_bl = m.subscribe("Current_Back_Left", 100, I_blCallback);
    ros::Subscriber I_br = m.subscribe("Current_Back_Right", 100, I_brCallback);

    ros::Subscriber battery = m.subscribe("Battery_Level", 100, batCallback);

    ros::Subscriber ctl_fl = m.subscribe("Control_Signal_Front_Left", 100, ctl_flCallback);
    ros::Subscriber ctl_fr = m.subscribe("Control_Signal_Front_Right", 100, ctl_frCallback);
    ros::Subscriber ctl_bl = m.subscribe("Control_Signal_Back_Left", 100, ctl_blCallback);
    ros::Subscriber ctl_br = m.subscribe("Control_Signal_Back_Right", 100, ctl_brCallback);

    ros::Subscriber vel = m.subscribe("cmd_vel",100,vel_Callback);

    //Declaration of the publisher that will tell 'base_controller' to launch the terrain test
    ros::Publisher pub = m.advertise<std_msgs::Bool>("runTest",0);
    std_msgs::Bool runTest;

    ros::Rate loop_rate(20);


    while (ros::ok()){

        float* Pw = computePowers();

        bool changeDetected = powerVariance(Pw);

        ros::spinOnce();
        loop_rate.sleep();

        //if a change is detected the message is published
        if (changeDetected){
            runTest.data = 1;
            pub.publish(runTest);
        }
    }

    return 0;
}
